/***
	Triworm.c, by Jacob Ewing
	Drags a cute little string of technicolour triangles behind the mouse.
***/
#include <stdlib.h>
#include <mem.h>
#include <math.h>

/*
	This structure is only used for drawing polygons.  It is defined here
	because polygon() passes it as an argument to compare_range(), which
	is a subcomponent of the polygon function.
*/
typedef struct{
	int x1, x2;
	char doneonce;
}rangetype;

/************************** function declarations ***************************/
char buttonpressed(void);
int mousex(void);
int mousey(void);
void initmouse(void);
char sgn(int val);
void initscrn(char mode);
unsigned char csmode(void);
void putpixel(int x, int y, unsigned char colour, char mode);
void line(int x1, int y1, int x2, int y2, unsigned char colour, char mode);
void compare_range(int x1, int y1, int x2, int y2, int miny, rangetype *range);
char polygon(int *x, int *y, int numpoints, unsigned char colour, char mode);
void hline(int x1, int x2, int y, unsigned char colour, char mode);
void palette(int c, char r, char g, char b);
void cls(void);

/************************** Global Variables ***************************/
unsigned char far *vmem = (unsigned char far *)0xA0000000;

/************************** function definitions ***************************/
void main(void){
	const int numpoints = 32, chunksize = 4;
	float *oldx, *oldy, *x, *y;
	int drawx[4], drawy[4];
	int a, b, n;
	float slope, dist;
	unsigned char smode;

	x = (float *)malloc(sizeof(float) * numpoints);
	y = (float *)malloc(sizeof(float) * numpoints);
	oldx = (float *)malloc(sizeof(float) * numpoints);
	oldy = (float *)malloc(sizeof(float) * numpoints);

	if(x && y && oldx && oldy){
		smode = csmode();
		initscrn(0x13);
		initmouse();
		a = mousex();
		b = mousey();
		for(n = 0; n < numpoints; n++){
			x[n] = a + n * (sqrt((chunksize * chunksize) / 2));
			y[n] = b + n * (sqrt((chunksize * chunksize) / 2));
		}
		for(n = 1; n < numpoints; n++){
			drawx[0] = x[n] + (y[n - 1] - y[n]) / 2;
			drawy[0] = y[n] - (x[n - 1] - x[n]) / 2;
			drawx[1] = x[n] - (y[n - 1] - y[n]) / 2;
			drawy[1] = y[n] + (x[n - 1] - x[n]) / 2;
			drawx[2] = x[n] + (x[n] - x[n - 1]) * 8;
			drawy[2] = y[n] + (y[n] - y[n - 1]) * 8;
			polygon(drawx, drawy, 3, 31 + n, 0);
		}

		while(buttonpressed());

		while(!buttonpressed()){
			if(x[0] != mousex() || y[0] != mousey()){
				oldx[0] = x[0];
				oldy[0] = y[0];
				x[0] = mousex();
				y[0] = mousey();

				for(n = 1; n < numpoints; n++){
					dist = sqrt(abs((x[n] - x[n - 1]) * (x[n] - x[n - 1]) +
									(y[n] - y[n - 1]) * (y[n] - y[n - 1])));
					if(dist != 0){
						slope = (float)chunksize / dist;
						oldy[n] = y[n];
						oldx[n] = x[n];
						x[n] = - slope * (x[n - 1] - x[n]) + x[n - 1];
						y[n] = - slope * (y[n - 1] - y[n]) + y[n - 1];
					}
				}
				for(n = 1; n < numpoints; n++){
					drawx[0] = oldx[n] + (oldy[n - 1] - oldy[n]) / 2;
					drawy[0] = oldy[n] - (oldx[n - 1] - oldx[n]) / 2;
					drawx[1] = oldx[n] - (oldy[n - 1] - oldy[n]) / 2;
					drawy[1] = oldy[n] + (oldx[n - 1] - oldx[n]) / 2;
					drawx[2] = oldx[n] + (oldx[n] - oldx[n - 1]) * 8;
					drawy[2] = oldy[n] + (oldy[n] - oldy[n - 1]) * 8;
					polygon(drawx, drawy, 3, 0, 0);

					drawx[0] = x[n] + (y[n - 1] - y[n]) / 2;
					drawy[0] = y[n] - (x[n - 1] - x[n]) / 2;
					drawx[1] = x[n] - (y[n - 1] - y[n]) / 2;
					drawy[1] = y[n] + (x[n - 1] - x[n]) / 2;
					drawx[2] = x[n] + (x[n] - x[n - 1]) * 8;
					drawy[2] = y[n] + (y[n] - y[n - 1]) * 8;
					polygon(drawx, drawy, 3, 31 + n, 0);
				}
			}
		}
		initscrn(smode);
		while(buttonpressed());
	}
	if(x) free(x);
	if(y) free(y);
	if(oldx) free(oldx);
	if(oldy) free(oldy);
}

/**************************** Mouse I/O *****************************/

/* returns X coordinate of mouse cursor position */
int mousex(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, cx
	}
	return returnval >> 1; /* The ">> 1" is for a screen mode that is 320
							  pixels wide.  When in this screen mode, this
							  interrupt returns a value from 0 to 639.
						   */
}

/* returns Y coordinate of mouse cursor position */
int mousey(void){
	int returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, dx
	}
	return returnval;

}

/* returns status of mouse butons. return value is laid out like so:

	76543210
	         left button (1 == pressed)
	        right button (1 == pressed)
	       middle button (1 == pressed)
	  unused
*/
char buttonpressed(void){
	char returnval;
	asm{
		mov ax, 0x3
		int 0x33
		mov returnval, bl
	}
	return returnval;
}

/* initializes the mouse */
void initmouse(void){
	asm{
		mov ax, 0
		int 0x33
	}
}

/****************************** Graphics ******************************/
/* initializes screen to specified mode */
void initscrn(char mode){
	asm{
		mov ah, 0
		mov al, mode
		int 0x10
	}
}

/* returns current screen mode */
unsigned char csmode(void){
	unsigned char returnval;
	asm{
		mov ah, 0xF
		int 0x10
		mov returnval, al
	}
	return returnval;
}

/* Draws a pixel at point (x, y). Pixels can be drawn in 7 different modes:

   mode:  Behavior:
  
	 0   Draw, no matter what
	 1   XOR the value onto the colour of the pixel already there.
	 2   OR the value onto the colour of the pixel already there.
	 3   AND the value onto the colour of the pixel already there.
	 4   Draw as long as the colour specified is unequal to zero.
	 5   Draw as long as the colour already there is equal to zero.
	 6   Add the value to the colour of the pixel already there.

	NOTE: drawing modes are the same for all other functions, with the
	exception of those which draw a specified shape.  In those cases,
	mode 4 draws an outline of the shape. eg. polygon( ... , 4);
*/
void putpixel(int x, int y, unsigned char colour, char mode){
	if((x >= 0) && (x <= 319) && (y >= 0) && (y <= 199))
		switch(mode){
			case 0: vmem[(y << 8) + (y << 6) + x]  = colour; break;
			case 1: vmem[(y << 8) + (y << 6) + x] ^= colour; break;
			case 2: vmem[(y << 8) + (y << 6) + x] |= colour; break;
			case 3: vmem[(y << 8) + (y << 6) + x] &= colour; break;
			case 4:
				if(colour) vmem[(y << 8) + (y << 6) + x] = colour;
				break;
			case 5:
				if(!vmem[(y << 8) + (y << 6) + x])
					vmem[(y << 8) + (y << 6) + x] = colour;
				break;
			case 6: vmem[(y << 8) + (y << 6) + x] += colour; break;
		}
}

/**
	Draws a line segment from point (x1, y1) to point (x2, y2).
	It has an odd layout, but doing it this way means it can be done strictly
	with integers.  This is, of course, far faster than using the classic
	"deltay = slope * deltax" method, which requires floating point math.
**/
void line(int x1, int y1, int x2, int y2, unsigned char colour, char mode){
	int n, deltax, deltay, sgndeltax, sgndeltay;
	int deltaxabs, deltayabs, x, y, drawx, drawy;

	deltax = x2 - x1;
	deltay = y2 - y1;
	deltaxabs = abs(deltax);
	deltayabs = abs(deltay);
	sgndeltax = sgn(deltax);
	sgndeltay = sgn(deltay);
	x = deltayabs >> 1;
	y = deltaxabs >> 1;
	drawx = x1;
	drawy = y1;

	putpixel(drawx, drawy, colour, mode);

	if(deltaxabs >= deltayabs){
		for(n = 0; n < deltaxabs; n++){
			y += deltayabs;
			if(y >= deltaxabs){
				y -= deltaxabs;
				drawy += sgndeltay;
			}
			drawx += sgndeltax;
			putpixel(drawx, drawy, colour, mode);
		}
	}else{
		for(n = 0; n < deltayabs; n++){
			x += deltaxabs;
			if(x >= deltayabs){
				x -= deltayabs;
				drawx += sgndeltax;
			}
			drawy += sgndeltay;
			putpixel(drawx, drawy, colour, mode);
		}
	}
}

/*
	This function is actually only a subcomponent of the polygon function.
	It tests each point on the line specified, comparing it with the
	corresponding values stored in range.  If the x component of the point
	is less than x1, then it's value is assigned to x1.  If the x component of the point
	is greater than x2, then it's value is assigned to x2.
*/
void compare_range(int x1, int y1, int x2, int y2, int miny, rangetype *range){
	int n, deltax, deltay, sgndeltax, sgndeltay, index;
	int deltaxabs, deltayabs, x, y, drawx, drawy;

	deltax = x2 - x1;
	deltay = y2 - y1;
	deltaxabs = abs(deltax);
	deltayabs = abs(deltay);
	sgndeltax = sgn(deltax);
	sgndeltay = sgn(deltay);
	x = deltayabs >> 1;
	y = deltaxabs >> 1;
	drawx = x1;
	drawy = y1;

	/***** compare current position with range ****/
	index = drawy - miny;
	if(range[index].doneonce){
		if(range[index].x1 > drawx)
			range[index].x1 = drawx;
		if(range[index].x2 < drawx)
			range[index].x2 = drawx;
	}else{
		range[index].x1 = drawx;
		range[index].x2 = drawx;
		range[index].doneonce = 1;
	}
	if(range[index].x1 < 0) range[index].x1 = 0;
	if(range[index].x2 < 0) range[index].x2 = 0;
	if(range[index].x1 > 319) range[index].x1 = 319;
	if(range[index].x2 > 319) range[index].x2 = 319;

	if(deltaxabs >= deltayabs){
		for(n = 0; n < deltaxabs; n++){
			y += deltayabs;
			if(y >= deltaxabs){
				y -= deltaxabs;
				drawy += sgndeltay;
			}
			drawx += sgndeltax;

			/***** compare current position with range ****/
			index = drawy - miny;
			if(range[index].doneonce){
				if(range[index].x1 > drawx)
					range[index].x1 = drawx;
				if(range[index].x2 < drawx)
					range[index].x2 = drawx;
			}else{
				range[index].x1 = drawx;
				range[index].x2 = drawx;
				range[index].doneonce = 1;
			}
			if(range[index].x1 < 0) range[index].x1 = 0;
			if(range[index].x2 < 0) range[index].x2 = 0;
			if(range[index].x1 > 319) range[index].x1 = 319;
			if(range[index].x2 > 319) range[index].x2 = 319;
		}
	}else{
		for(n = 0; n < deltayabs; n++){
			x += deltaxabs;
			if(x >= deltayabs){
				x -= deltayabs;
				drawx += sgndeltax;
			}
			drawy += sgndeltay;

			/***** compare current position with range ****/
			index = drawy - miny;
			if(range[index].doneonce){
				if(range[index].x1 > drawx)
					range[index].x1 = drawx;
				if(range[index].x2 < drawx)
					range[index].x2 = drawx;
			}else{
				range[index].x1 = drawx;
				range[index].x2 = drawx;
				range[index].doneonce = 1;
			}
			if(range[index].x1 < 0) range[index].x1 = 0;
			if(range[index].x2 < 0) range[index].x2 = 0;
			if(range[index].x1 > 319) range[index].x1 = 319;
			if(range[index].x2 > 319) range[index].x2 = 319;
		}
	}
}

/*
	draws a polygon defined by the lines which sequentially connect the points
	specified int *x and *y.
	IMPORTANT:  This function is guaranteed to work if, and only if, no two
	points in the polygon can define a line segment that goes outside the
	polygon.  This is a bit of hinderance, but I think it's forgivable since
	it cuts down on the processing time.
*/
char polygon(int *x, int *y, int numpoints, unsigned char colour, char mode){
	rangetype *range;
	int n, miny, maxy, ysize;
	char returnval = 0;

	if(mode == 4){
		for(n = 0; n < numpoints; n++)
			line(x[n], y[n], x[(n + 1) % numpoints], y[(n + 1) % numpoints],
				 colour, 0);
	}else{

		miny = maxy = y[0];
		for(n = 1; n < numpoints; n++){
			if(y[n] < miny) miny = y[n];
			if(y[n] > maxy) maxy = y[n];
		}
		ysize = maxy - miny + 1;

		range = (rangetype *)malloc(ysize * sizeof(rangetype));
		if(range){
			/** initialize ranges **/
			for(n = 0; n < ysize; n++) range[n].doneonce = 0;

			/** calculate ranges **/
			for(n = 0; n < numpoints; n++){
				compare_range(x[n], y[n], x[(n + 1) % numpoints],
							  y[(n + 1) % numpoints], miny, range);
			}

			/** draw polygon **/
			for(n = miny; n <= maxy; n++){
				if(n >= 0 && n < 200)
				hline(range[n - miny].x1, range[n - miny].x2,
					  n, colour, mode);
			}
			free(range);
		}else returnval = 1;
	}
	return returnval;
}

/* draws a horizontal line on the screen */
void hline(int x1, int x2, int y, unsigned char colour, char mode){
	/* WARNING this function assumes (when drawing in mode 0) that
	   a range check has already been done.
	*/
	int n;
	if(!mode)
		_fmemset(&vmem[x1 + 320 * y], colour, x2 - x1 + 1);
	else
		for(n = x1; n <= x2; n++)
			putpixel(n, y, colour, mode);
}

/* adjusts actual colour of specified colour index */
void palette(int c, char r, char g, char b){
	r %= 64;
	g %= 64;
	b %= 64;
	asm{
		mov ah, 0x10
		mov al, 0x10
		mov bx, c
		mov ch, g
		mov cl, b
		mov dh, r
		int 0x10
	}
}

/* clears the screen when in VGA 320 * 200 * 8 */
void cls(void){
	_fmemset(vmem, 0, 64000);
}

/****************************** Misc ******************************/

/* returns the sign of a number
	val:		return value:
 
	< 0    		-1
	> 0    		 1
	= 0             0

*/
char sgn(int val){
	if(val > 0) return 1;
	return -(val < 0);
}
